Iconifier = class(Turbine.UI.Window);

function Iconifier:Constructor()
    Turbine.UI.Window.Constructor(self);
    
    AddCallback(Turbine.UI.Display, "SizeChanged", function()
        self:DisplaySizeChanged();
    end);
end

function Iconifier:SetLeft(left)
    self:SetPosition(left, self:GetTop());
end

function Iconifier:SetTop(top)
    self:SetPosition(self:GetLeft(), top);
end

function Iconifier:SetPosition(left, top)
    Turbine.UI.Window.SetPosition(self, left, top);
    local screenWidth, screenHeight = Turbine.UI.Display:GetSize();
    local center = left + self:GetWidth() / 2;
    local middle = top + self:GetHeight() / 2;
    self.x = center / screenWidth;
    self.y = middle / screenHeight;
end

function Iconifier:DisplaySizeChanged()
    self:SetFractionalPosition(self.x, self.y);
end

function Iconifier:SetFractionalPosition(x, y)
    self.x, self.y = x, y;
    local screenWidth, screenHeight = Turbine.UI.Display:GetSize();
    local center = screenWidth * x;
    local middle = screenHeight * y;
    local left = center - self:GetWidth() / 2;
    local top = middle - self:GetHeight() / 2;
    Turbine.UI.Window.SetPosition(self, math.floor(0.5 + left), math.floor(0.5 + top));
end

function Iconifier:GetFractionalPosition()
    return self.x, self.y;
end

function Iconifier:SetIcon(ctl)
    if (self.icon) then
        self.icon:SetParent(nil);
    end
    self.icon = ctl;
    ctl:SetParent(self);
    AddCallback(ctl, "MouseClick", function(_, args) self:MouseClick(args) end);
    AddCallback(ctl, "MouseDown",  function(_, args) self:MouseDown(args)  end);
    AddCallback(ctl, "MouseMove",  function(_, args) self:MouseMove(args)  end);
    AddCallback(ctl, "MouseUp",    function(_, args) self:MouseUp(args)    end);
    AddCallback(ctl, "MouseHover", function(_, args) self:MouseHover(args) end);
    AddCallback(ctl, "MouseLeave", function(_, args) self:MouseLeave(args) end);
    AddCallback(ctl, "SizeChanged", function()
        self:SetSize(ctl:GetSize());
    end);
    self:SetSize(ctl:GetSize());
end

function Iconifier:SetWindow(win)
    self.window = win;
    
    AddCallback(win, "Closing", function(_, args)
        self:Iconify();
        args.Cancel = true;
    end);
end

function Iconifier:Iconify()
    self.window:SetVisible(false);
    self.zoomingTo = "icon";
    self.zoomStartTime = Turbine.Engine.GetGameTime();
    self:SetWantsUpdates(true);
end

function Iconifier:Restore()
    self:SetVisible(false);
    self.zoomingTo = "window";
    self.zoomStartTime = Turbine.Engine.GetGameTime();
    self:SetWantsUpdates(true);
end

function Iconifier:MouseClick(args)
    if ((args.Button == Turbine.UI.MouseButton.Left) and not self.moved) then
        self:Restore();
    end
end
    
function Iconifier:MouseDown(args)
    if (args.Button == Turbine.UI.MouseButton.Left) then
        self.mouseDown = { Turbine.UI.Display:GetMousePosition() };
        self.moved = nil;
    end
end

function Iconifier:MouseMove()
    if (self.mouseDown) then
        self.moved = true;
        local oldX, oldY = unpack(self.mouseDown);
        local newX, newY = Turbine.UI.Display:GetMousePosition();
        local posX, posY = self:GetPosition();
        self:SetPosition(posX + newX - oldX, posY + newY - oldY);
        self.mouseDown = { newX, newY };
    end
end

function Iconifier:MouseUp(args)
    if (args.Button == Turbine.UI.MouseButton.Left) then
        if (self.moved) then
            DoCallbacks(self, "IconMoved");
        end
        self.mouseDown = nil;
    end
end

function Iconifier:Update()
    local progress = (Turbine.Engine.GetGameTime() - self.zoomStartTime) * 2;
    local param = progress^3;
    if (self.zoomingTo == "icon") then
        param = 1 - param;
    end
    if (not self.zoomIndicator) then
        self.zoomIndicator = Turbine.UI.Window();
        self.zoomIndicator:SetBackColor(Turbine.UI.Color(1, 0.80, 0.80, 0.80));
        self.zoomIndicator:SetVisible(true);
        local innerCtl = Turbine.UI.Control();
        innerCtl:SetBackColor(Turbine.UI.Color(0, 0, 0 ,0));
        innerCtl:SetParent(self.zoomIndicator);
        innerCtl:SetPosition(2, 2);
        self.zoomIndicator.SizeChanged = function(z)
            innerCtl:SetWidth(z:GetWidth() - 4);
            innerCtl:SetHeight(z:GetHeight() - 4);
        end
    end
    if (progress >= 1) then
        self.zoomIndicator:SetVisible(false);
        self.zoomIndicator = nil;
        self:SetIconState(self.zoomingTo);
        self:SetWantsUpdates(false);
        DoCallbacks(self, "IconStateChanged", self.zoomingTo);
    else
        local winLeft, winTop = self.window:GetPosition();
        local winWidth, winHeight = self.window:GetSize();
        local iconLeft, iconTop = self:GetPosition();
        local iconWidth, iconHeight = self:GetSize();
        self.zoomIndicator:SetLeft(iconLeft + param * (winLeft - iconLeft));
        self.zoomIndicator:SetTop(iconTop + param * (winTop - iconTop));
        self.zoomIndicator:SetWidth(iconWidth + param * (winWidth - iconWidth));
        self.zoomIndicator:SetHeight(iconHeight + param * (winHeight - iconHeight));
    end
end

function Iconifier:GetIconState()
    if (self:IsVisible()) then
        return "icon";
    else
        return "window";
    end
end

function Iconifier:IsIcon()
    return self:IsVisible();
end

function Iconifier:IsWindow()
    return not self:IsIcon();
end

function Iconifier:SetIconState(state)
    self:SetVisible(false);
    self.window:SetVisible(false);
    if (state == "icon") then
        self:SetVisible(true);
    elseif (state == "window") then
        self.window:SetVisible(true);
        self.window:Activate();
    end
end

Thurallor = Thurallor or {};
Thurallor.UI = Thurallor.UI or {};
Thurallor.UI.Iconifier = Iconifier;
