TableControl = class(Turbine.UI.Control);

-- Add SetRowVisible().
-- edgemover can't be moved too far, set min. column width
-- SetColumnVisible(), add right-click menu to select columns
-- Derive each column from TreeView().  Use :SetFilter() for row visibility.  Use :Sort().
-- Add horizontal / vertical orientation?
-- MinimumSize (horizontal) style

function TableControl:Constructor(style)
    Turbine.UI.Control.Constructor(self);
    self:SetMouseVisible(false);
 
    if (not style) then
        style = {}
        style.horizontalCellSpacing = -1;
        style.verticalCellSpacing = -1;
        style.minimumColumnWidth = 10;
        style.heading = {
    
            -- Built-in styles (Turbine.UI.Control)
            BackColor = Turbine.UI.Color(1, 0, 0.38, .67);
            Enabled = true;
            Font = Turbine.UI.Lotro.Font.Verdana14;
            ForeColor = Turbine.UI.Color(1, 1, 1, 1);
            Height = 24;
            Multiline = false;
            TextAlignment = Turbine.UI.ContentAlignment.MiddleCenter;
            ReadOnly = true;
            Selectable = false;
            MarkupEnabled = true; -- so we can underline the text when sorting

            -- Special styles (TableControl)
            Object = Turbine.UI.Lotro.TextBox; -- what kind of object to create for each heading
            HighlightColor = Turbine.UI.Color(1, 0, 0.57, 1);
        };
        style.column = {
            SortMethod = function(a, b)
                -- Takes two cells as arguments, returns true/false
                return (a:GetText() < b:GetText());
            end
        }
        style.cell = {

            -- Built-in styles (Turbine.UI.Control)
            BackColor = Turbine.UI.Color(1, 0, 0.19, 0.33);
            Enabled = true;
            Font = Turbine.UI.Lotro.Font.Verdana14;
            ForeColor = Turbine.UI.Color(1, 0.75, 0.75, 0.75);
            Height = 24;
            Multiline = false;
            TextAlignment = Turbine.UI.ContentAlignment.MiddleCenter;
            ReadOnly = true;
            
            -- Special styles (TableControl)
            Object = Turbine.UI.Lotro.TextBox; -- what kind of object to create for each cell
            HighlightColor = Turbine.UI.Color(1, 0, 0.57, 1);
        };
        style.edgeMover = {
        
            -- Built-in styles (Turbine.UI.Control)
            Width = 5;
            BackColor = Turbine.UI.Color(0.75, 0, .19, .33);
            BackColorBlendMode = Turbine.UI.BlendMode.Overlay;
        };
        style.marquee = {
            Color = Turbine.UI.Color.White;
            Speed = 20; -- updates per second
        }
    end

    self.rows = Turbine.UI.ListBox();
    self.rowsByName = {};
    self.rows:SetParent(self);
    -- The cell container must be in front of the headings, in case they overlap
    -- due to verticalCellSpacing < 0.
    self.rows:SetZOrder(1);
    self.rows.MouseClick = function()
        self:Deselect();
    end
    self.rows.SelectedIndexChanged = function()
        self:_SelectRow(self.rows:GetSelectedItem());
    end
    self.rowWidth = 0;
    
    self.vScrollBar = Turbine.UI.Lotro.ScrollBar();
    self.vScrollBar:SetOrientation(Turbine.UI.Orientation.Vertical);
    self.vScrollBar:SetParent(self);
    self.vScrollBar:SetWidth(10);
    self.rows:SetVerticalScrollBar(self.vScrollBar);
    self.vScrollBar.VisibleChanged = function()
        self:SetWidth(self:GetWidth());
    end

    self.columns = {};
    self.columnsByName = {};
    self.totalWeight = 0;
    self.edgeMovers = {};

    self:SetStyle(style);
 end

function TableControl:SetStyle(style)
    self.style = style;
    self.rows:SetTop(self.style.heading.Height + self.style.verticalCellSpacing);
end

function TableControl:GetDimensions()
    return #self.rows:GetItemCount(), #self.columns;
end

function TableControl:AddColumn(name, text, weight, sortMethod)
    if (weight == nil) then
        weight = 1.0;
    end
    if (sortMethod == nil) then
        sortMethod = self.style.column.SortMethod;
    end

    -- Create and initialize new column object
    local column = {
        name = name;
        heading = nil;
        weight = 0; -- this gets set below
        sortMethod = sortMethod;
    };
    table.insert(self.columns, column);
    local col = #self.columns;
    self.columnsByName[name] = col;

    -- Add a cell to each row and a heading
    column.heading = self:_CreateHeading(col, text);
    for row = 1, self.rows:GetItemCount() do
        self:SetCell(row, col);
    end
    
    self:SetColumnWeight(col, weight); -- also calls _UpdateColumnWidths()
    
    return column;
end

function TableControl:_CreateHeading(col, text)
    local heading = self.style.heading.Object();
    self:_ApplyStyles(heading, self.style.heading);
    heading:SetParent(self);
    heading:SetText(text);
    heading.col = col;

    AddCallback(heading, "MouseEnter", function()
        heading.mouseInside = true;
        heading:SetBackColor(self.style.heading.HighlightColor);
    end);
    AddCallback(heading, "MouseLeave", function()
        heading.mouseInside = false;
        if (not heading.mouseDown) then
            heading:SetBackColor(self.style.heading.BackColor);
        end
    end);
    AddCallback(heading, "MouseDown", function(_, args)
        if (args.Button == Turbine.UI.MouseButton.Left) then
            heading.mouseDown = true;
            heading.mouseStartX = Turbine.UI.Display.GetMouseX();
            heading.startLeft = heading:GetLeft();
        end
    end);
    AddCallback(heading, "MouseClick", function(_, args)
        if (args.Button == Turbine.UI.MouseButton.Left) then
            if (heading.mouseMoved) then
                heading.mouseMoved = false;
            else
                self:_SortColumn(heading.col);
            end
        elseif (args.Button == Turbine.UI.MouseButton.Right) then
            self:_ShowHeadingMenu(heading.col);
        end
    end);
    AddCallback(heading, "MouseMove", function(_, args)
        if (heading.mouseDown) then
            heading.mouseMoved = true;
            heading:SetZOrder(1);
            local deltaX = Turbine.UI.Display.GetMouseX() - heading.mouseStartX;
            local left = math.min(math.max(0, heading.startLeft + deltaX), self.rowWidth - heading:GetWidth());
            heading:SetLeft(left);
            deltaX = left - heading.startLeft;
            self:_ColumnDragged(heading.col, deltaX);
        end
    end);
    AddCallback(heading, "MouseUp", function(_, args)
        if (args.Button == Turbine.UI.MouseButton.Left) then
            heading.mouseDown = false;
            if (heading.mouseMoved) then
                heading:SetZOrder(0);
                self:_UpdateColumnWidths();
            end
            if (not heading.mouseInside) then
                heading:SetBackColor(self.style.heading.BackColor);
            end
        end
    end);

    return heading;
end

function TableControl:_ColumnDragged(col, deltaX)
    local heading = self.columns[col].heading;
    local leftEdge = heading:GetLeft();
    local rightEdge = leftEdge + heading:GetWidth();

    local swapped = false;
    if (deltaX < 0) then
        -- Find visible column to the left, if any
        for c = col - 1, 1, -1 do
            local other = self.columns[c];
            if (other.weight > 0) then
                local center = other.heading:GetLeft() + math.floor(0.5 + (other.heading:GetWidth() / 2));
                if (leftEdge < center) then
--Puts("deltaX = " .. deltaX .. "; halfWidth = " .. halfWidth .. "; hw = " .. hw);
--Puts("swapping left: " .. col .. ", " .. c);                
                    self:_SwapColumns(col, c);
                    swapped = true;
                    break;
                end
            end
        end
    else
        -- Find visible column to the right, if any
        for c = col + 1, #self.columns do
            local other = self.columns[c];
            if (other.weight > 0) then
                local center = other.heading:GetLeft() + math.floor(0.5 + (other.heading:GetWidth() / 2));
                if (rightEdge > center) then
--Puts("deltaX = " .. deltaX .. "; halfWidth = " .. halfWidth .. "; hw = " .. hw);
--Puts("swapping right: " .. col .. ", " .. c);                
                    self:_SwapColumns(col, c);
                    swapped = true;
                    break;
                end
            end
        end
    end
    
--Puts("before: " .. heading:GetLeft());      
    heading.startLeft = heading:GetLeft();
    heading:SetLeft(leftEdge);
    deltaX = leftEdge - heading.startLeft;
    heading.mouseStartX = Turbine.UI.Display:GetMouseX() - deltaX;
--Puts("after: " .. heading:GetLeft());                    
end

function TableControl:_SwapColumns(a, b)
    local temp = self.columns[a];
    self.columns[a] = self.columns[b];
    self.columns[b] = temp;
    
    self.columns[a].heading.col = a;
    self.columns[b].heading.col = b;
    self.columnsByName[self.columns[a].name] = a;
    self.columnsByName[self.columns[b].name] = b;
    
    for row = 1, self.rows:GetItemCount() do
        local cells = self.rows:GetItem(row).cells;
        temp = cells[a];
        cells[a] = cells[b];
        cells[b] = temp;
    end
    
    self:_UpdateColumnWidths();
    self:_DoCallbacks("PresentationChanged", { Desc = "Columns swapped positions" });
end

function TableControl:_ShowHeadingMenu(col)
    local column = self.columns[col];
    local contextMenu = Turbine.UI.ContextMenu();
    local menuItems = contextMenu:GetItems();

    if (self.totalWeight > column.weight) then
        local hideItem = Turbine.UI.MenuItem("Hide this column");
        hideItem.Click = function()
            self:SetColumnWeight(col, 0);
        end
        menuItems:Add(hideItem);
    end

    if (column.sortMethod) then
        local sortItem = Turbine.UI.MenuItem("Sort by this column");
        sortItem.Click = function()
            self:_SortColumn(col);
        end
        menuItems:Add(sortItem);
    end

    if (#self.columns > 1) then
        local showItem = Turbine.UI.MenuItem("Show columns");
        local colItems = showItem:GetItems();
        for c = 1, #self.columns do
            local other = self.columns[c];
            local name = other.heading:GetText();
            local item = Turbine.UI.MenuItem(name, true, true);
            if (other.weight > 0) then
                item.Click = function()
                    self:SetColumnWeight(c, 0);
                end
            else
                item:SetChecked(false);
                item.Click = function()
                    self:SetColumnWeight(c, 1);
                end
            end
            colItems:Add(item);
        end
        menuItems:Add(showItem);
    end

    contextMenu:ShowMenu();
end

function TableControl:SortColumn(name, dir)
    local col = self.columnsByName[name];
    local column = self.columns[col];
    column.sortDirection = not dir;
    self:_SortColumn(col);
end

function TableControl:_SortColumn(col)
    local column = self.columns[col];
    if (not column.sortMethod) then
        return;
    end
    
    -- Show sort indication icon
    if (not self.sortIcon) then
        self.sortIcon = Turbine.UI.Label();
        self.sortIcon:SetFont(Turbine.UI.Lotro.Font.Arial12);
        self.sortIcon:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
        self.sortIcon:SetMouseVisible(false);
        self.sortIcon:SetFontStyle(Turbine.UI.FontStyle.Outline);
        self.sortIcon:SetOutlineColor(Turbine.UI.Color(0.75, 0, 0, 0));
    end;
    self.sortIcon:SetParent(column.heading);
    self.sortIcon:SetZOrder(1);
    self.sortIcon:SetSize(15, column.heading:GetHeight());
    if (column.sortDirection) then
        self.sortIcon:SetText("▼");
    else
        self.sortIcon:SetText("▲");
    end

    -- Toggle between ascending/descending when creating the sort function
    column.sortDirection = not column.sortDirection;
    local compare;
    if (column.sortDirection) then
        compare = function(a, b)
            -- Takes rowContainers as args, returns true or false
            return column.sortMethod(a.cells[col], b.cells[col]);
        end
    else
        compare = function(a, b)
            -- Takes rowContainers as args, returns true or false
            return column.sortMethod(b.cells[col], a.cells[col]);
        end
    end

    self.lastSort = { col = col; compare = compare };
    self:_ReSort();
    
    self:_DoCallbacks("PresentationChanged", { Desc = "Column sort changed" });
end

function TableControl:_CreateCell(row, col)
    local rowContainer = self.rows:GetItem(row);
    local cell = self.style.cell.Object();
    self:_ApplyStyles(cell, self.style.cell);
    return cell;
end

function TableControl:_CreateEdgeMovers()
    -- Destroy previous objects
    for edge = 1, #self.edgeMovers do
        self.edgeMovers[edge]:SetParent(nil);
    end
    self.edgeMovers = {};

    -- Only show an EdgeMover between two visible (weight > 0) columns.
    local prev = nil;
    for col = 1, #self.columns do
        if (self.columns[col].weight > 0) then
            if (prev) then
                -- Found a pair of visible columns.
                local edgeMover = Thurallor.UI.EdgeMover(Turbine.UI.Orientation.Vertical);
                edgeMover.leftColumn = prev;
                edgeMover.rightColumn = col;
                edgeMover:SetParent(self);
                edgeMover:SetZOrder(1); -- in front of rows
                self:_ApplyStyles(edgeMover, self.style.edgeMover);
                AddCallback(edgeMover, "EdgeMoved", function()
                    self:_EdgeMoved(edgeMover);
                end);
                AddCallback(edgeMover, "MouseUp", function()
                    -- EdgeMover may set a column weight to zero; if so, disable the column's EdgeMover.
                    self:_CreateEdgeMovers();
                end);
                table.insert(self.edgeMovers, edgeMover);
            end
            prev = col;
        end
    end
    self:_PlaceEdgeMovers();
    self:_DoCallbacks("PresentationChanged", { Desc = "Column weight(s) changed" });
end

function TableControl:_EdgeMoved(edgeMover)
    local x, y = self:GetMousePosition();
    local minWidth = self.style.minimumColumnWidth;

    local leftCol = self.columns[edgeMover.leftColumn];
    local leftHeading = leftCol.heading;
    x = math.max(x, leftHeading:GetLeft() + minWidth);

    local rightCol = self.columns[edgeMover.rightColumn];
    local rightHeading = rightCol.heading;
    x = math.min(x, rightHeading:GetLeft() + rightHeading:GetWidth() - minWidth);
    
    local leftWidth = x - leftHeading:GetLeft();
    local rightWidth = rightHeading:GetLeft() + rightHeading:GetWidth() - x;
    local bothWidth = leftWidth + rightWidth;
    local bothWeight = leftCol.weight + rightCol.weight;

    leftCol.weight = bothWeight * (leftWidth / bothWidth);
    rightCol.weight = bothWeight * (rightWidth / bothWidth);

    self:_UpdateColumnWidths();
end

function TableControl:_DoCallbacks(event, args)
    if (not self.callbacksDisabled) then
        DoCallbacks(self, event, args);
    end
end

function TableControl:SetColumnWeight(col, weight)
    local column = self.columns[col];
    local prevWeight = column.weight;
    self.totalWeight = self.totalWeight - prevWeight + weight;
    column.weight = weight;
    if ((weight == 0) and (prevWeight > 0)) then
        column.heading:SetParent(nil);
        for row = 1, self.rows:GetItemCount() do
            local rowContainer = self.rows:GetItem(row);
            rowContainer.cells[col]:SetParent(nil);
        end
        self:_CreateEdgeMovers();
    elseif ((weight > 0) and (prevWeight == 0)) then
        column.heading:SetParent(self);
        for row = 1, self.rows:GetItemCount() do
            local rowContainer = self.rows:GetItem(row);
            rowContainer.cells[col]:SetParent(rowContainer);
        end
        self:_CreateEdgeMovers();
    end
    self:_UpdateColumnWidths();
    column.prevWeight = prevWeight;
end

-- "Presentation" includes (so far)
--   - the user's choices of column weights, visibility and ordering
--   - the sorted column and direction
function TableControl:GetPresentation()
    local columns = {};
    for col = 1, #self.columns do
        local column = self.columns[col];
        table.insert(columns, { column.name, column.weight });
    end
    
    local presentation = {};
    presentation.columns = columns;
    if (self.lastSort) then
        local column = self.columns[self.lastSort.col];
        presentation.sort = { column = column.name; dir = column.sortDirection };
    end
    
    return presentation;
end

function TableControl:SetPresentation(presentation)

    -- Get column weights from presentation
    local foundColumns = {};
    for col = 1, #presentation.columns do
        local name, weight = unpack(presentation.columns[col]);
        foundColumns[name] = weight;
    end

    -- Handle columns that aren't specified in the presentation
    --   Set their weight to 0.
    for name, col in pairs(self.columnsByName) do
        if (foundColumns[name] == nil) then
            foundColumns[name] = true;
            table.insert(presentation.columns, { name, 0 });
        end
    end

    -- Create a new cell list in each rowContainer
    for row = 1, self.rows:GetItemCount() do
        self.rows:GetItem(row).newCells = {};
    end

    -- Apply the new ordering and weights to the columns
    local oldColumns = self.columns;
    self.columns = {};
    self.totalWeight = 0;
    for col = 1, #presentation.columns do
        local name, weight = unpack(presentation.columns[col]);
        local oldCol = self.columnsByName[name];
        local column = oldColumns[oldCol];
        self.totalWeight = self.totalWeight + weight;
        column.weight = weight;
        column.heading.col = col;
        column.heading:SetParent((weight > 0) and self or nil);
        self.columnsByName[name] = col;
        table.insert(self.columns, column);
        for row = 1, self.rows:GetItemCount() do
            local rowContainer = self.rows:GetItem(row);
            local cell = rowContainer.cells[oldCol];
            rowContainer.newCells[col] = cell;
            cell:SetParent((weight > 0) and rowContainer or nil);
        end
    end
        
    -- Start using the new cell list in each rowContainer
    for row = 1, self.rows:GetItemCount() do
        local rowContainer = self.rows:GetItem(row);
        rowContainer.cells = rowContainer.newCells;
        rowContainer.newCells = {};
    end

    -- Move the cells into their new positions
    self:_CreateEdgeMovers();
    self:_UpdateColumnWidths();
    
    -- Apply the sort (if any)
    if (presentation.sort) then
        local col = self.columnsByName[presentation.sort.column];
        local column = self.columns[col];
        column.sortDirection = not presentation.sort.dir;
        self:_SortColumn(col);
    end
end

function TableControl:GetRow(name)
    
end

function TableControl:AddRow(name, moreRows)
    local rowContainer = Turbine.UI.Control();
    rowContainer:SetWidth(self.rowWidth);
    rowContainer.name = name;
    rowContainer.cells = {};
    rowContainer.maxCellHeight = 0;
    self.rows:AddItem(rowContainer);
    self.rowsByName[name] = rowContainer;
    local row = self.rows:GetItemCount();

    -- Add new cells
    self:SetRowVisible(row, false);
    for col = 1, #self.columns do
        self:SetCell(row, col);
    end
    
    -- Add mask overlay.  We want the user to have to select the row before clicking again
    -- on a particular cell to edit it.
    local mask = Turbine.UI.Control();
    mask:SetParent(rowContainer);
    mask:SetSize(rowContainer:GetSize());
    mask:SetZOrder(3); -- in front of cells
    rowContainer.mask = mask;

    -- Add event behaviors
    rowContainer.PositionChanged = function(ctl)
        if ((self.selectedRow == ctl) and self.marquee) then
            self.marquee:SetPosition(ctl:GetPosition());
        end
    end
    rowContainer.SizeChanged = function(ctl)
        ctl.mask:SetSize(ctl:GetSize());
        if ((self.selectedRow == ctl) and self.marquee) then
            self.marquee:SetSize(ctl:GetSize());
        end
    end
    mask.MouseEnter = function()
        for c = 1, #self.columns do
            rowContainer.cells[c]:SetBackColor(self.style.cell.HighlightColor);
        end
    end
    mask.MouseLeave = function()
        for c = 1, #self.columns do
            rowContainer.cells[c]:SetBackColor(self.style.cell.BackColor);
        end
    end
    mask.MouseClick = function(_, args)
        if (args.Button == Turbine.UI.MouseButton.Right) then
            DoCallbacks(rowContainer, "ContextMenu");
        end
    end
    
    self:_UpdateLastRow(); -- also calls SetRowVisible(row, true)

    -- Table needs to be resorted, if enabled
    if ((not moreRows) and self.lastSort) then
        self:ReSort();
-- if not auto-sorting, then hide the sort icon        
    end
    
    -- Return a table of cells indexed by column name
    local cells = {};
    for c = 1, #self.columns do
        local colName = self.columns[c].name;
        cells[colName] = rowContainer.cells[c];
    end
    return rowContainer, cells;
end

function TableControl:ReSort()
    self.lastSort.sorted = false;
    self:SetWantsUpdates(true);
end

function TableControl:Update()
    self:_ReSort();
    self:SetWantsUpdates(false);
end

-- Reapply the most recent sort
function TableControl:_ReSort()
    if (self.lastSort) then
        self.rows:Sort(self.lastSort.compare);
        self.lastSort.sorted = true;
        self:_UpdateLastRow();
    end
end

-- Can specify colName or nil; returns column name of sorted column or nil
function TableControl:IsSorted(colName)
    if ((self.lastSort) and (self.lastSort.sorted)) then
        local sortedCol = self.columns[self.lastSort.col];
        if (not colName) then
            return sortedCol.name;
        else
            if (self.lastSort.col == self.columnsByName[colName]) then
                return colName;
            end
        end
    end
end

function TableControl:SelectRow(rowName)
    if (rowname) then
        self:_SelectRow(self.rowsByName[rowName]);
    else
        self:_SelectRow(nil);
    end
end

function TableControl:_SelectRow(rowContainer)
    local oldSelectedRow = self.selectedRow;

    -- Re-enable mask on previous row
    if (oldSelectedRow) then
        oldSelectedRow.mask:SetParent(oldSelectedRow);
    end
    
    -- Remove old marquee, if any
    if (self.marquee) then
        self.marquee:SetParent(nil);
        self.marquee = nil;
    end

    if (rowContainer) then
        rowContainer:Focus();
    
        -- Create new marquee and attach it
        self.marquee = Thurallor.UI.Marquee();
        self:_ApplyStyles(self.marquee, self.style.marquee);
        local height = 0;
        for _, cell in pairs(rowContainer.cells) do
            height = math.max(height, cell:GetHeight());
        end
        self.marquee:SetPosition(rowContainer:GetPosition());
        self.marquee:SetSize(rowContainer:GetWidth(), height);
        self.marquee:SetParent(self.rows);
        
        -- Disable mask to allow mouse clicks to reach the cells
        rowContainer.mask:SetParent(nil);
    else
        self.rows:Focus();
    end
    
    self.selectedRow = rowContainer;
    local oldName = (oldSelectedRow and oldSelectedRow.name) or nil;
    local newName = (rowContainer and rowContainer.name) or nil;
    self:_DoCallbacks("SelectionChanged", { New = newName; Old = oldName });
end

function TableControl:Deselect()
    self:_SelectRow(nil);
end

function TableControl:InsertRows(beforeRow, numNewRows)
end

function TableControl:RowIsVisible(row)
    return (self.rows:GetItem(row):GetHeight() > 0);
end

function TableControl:EnsureVisible(rowName)
    if (rowName) then
        local rowContainer = self.rowsByName[rowName];
        self.rows:EnsureVisible(self.rows:IndexOfItem(rowContainer));
    end
end

function TableControl:SetRowVisible(row, visible)
    local rowContainer = self.rows:GetItem(row);
    if (visible) then
        -- Set container height to maximum of cell heights
        local maxHeight = 0;
        for _, cell in pairs(rowContainer.cells) do
            maxHeight = math.max(maxHeight, cell:GetHeight());
        end
        if (rowContainer ~= self.lastRow) then
            maxHeight = maxHeight + self.style.verticalCellSpacing;
        end
        rowContainer:SetHeight(maxHeight);
    else
        rowContainer:SetHeight(0);
    end
end

-- Last row needs to be slightly different cosmetically
function TableControl:_UpdateLastRow()
    local prev = self.lastRow;
    local row = self.rows:GetItemCount();
    if (row > 0) then
        self.lastRow = self.rows:GetItem(row);
        self:SetRowVisible(row, true);
    else
        self.lastRow = nil;
    end
    if (prev) then
        self:SetRowVisible(self.rows:IndexOfItem(prev), true);
    end
end

function TableControl:DeleteRow(rowName)
    local rowContainer = self.rowsByName[rowName];
    if (self.selectedRow == rowContainer) then
        self:Deselect();
    end
    self.rows:RemoveItem(rowContainer);
    self.rowsByName[rowName] = nil;
    if (self.lastRow == rowContainer) then
        self.lastRow = nil;
        self:_UpdateLastRow();
    end
end

function TableControl:GetRowCells(row)
    return unpack(self.rows:GetItem(row).cells);
end

function TableControl:GetCell(row, col)
    return self.rows:GetItem(row).cells[col];
end

function TableControl:SetCell(row, col, object)
    local column = self.columns[col];
    local heading = column.heading;
    local rowContainer = self.rows:GetItem(row);

    -- Delete old cell (if any)
    local oldCell = rowContainer.cells[col];
    if (oldCell) then
        oldCell:SetParent(nil);
        if (oldCell.Close) then
            oldCell:Close();
        end
    end
    
    -- Create new cell, or use the supplied object
    local newCell = object;
    if (object == nil) then
        newCell = self:_CreateCell(row, col);
    end

    -- Place the cell within in the row container
    local left = heading:GetLeft();
    local width = heading:GetWidth();
    newCell:SetParent(rowContainer);
    if (column.weight == 0) then
        newCell:SetParent(nil);
    end
    rowContainer.cells[col] = newCell;
    newCell:SetWidth(width);
    newCell:SetPosition(left, top);
    
    -- Update the row container height, if necessary
    if (rowContainer:GetHeight() > 0) then
        self:SetRowVisible(true);
    end

    return newCell;
end

function TableControl:_ApplyStyles(object, styles)
    for s in keys(styles) do
        local SetFunction = object["Set" .. s];
        if (SetFunction) then
            SetFunction(object, styles[s]);
        end
    end
end

function TableControl:_UpdateColumnWidths()
    -- Find last visible column
    local rightCol;
    for col = 1, #self.columns do
        local column = self.columns[col];
        if (column.weight > 0) then
            rightCol = col;
        end
    end

    local totalWidth = self.rowWidth - ((#self.columns - 1) * self.style.horizontalCellSpacing);
    local totalWeight = self.totalWeight;
    local left = 0;
    local weightExpended = 0;
    for col = 1, #self.columns do
        local column = self.columns[col];
        local weight = column.weight;
        if (weight > 0) then
            local heading = column.heading;
            local width = math.floor(0.5 + totalWidth * (weight / totalWeight));
            width = math.min(width, self.rowWidth - left)
            if (col == rightCol) then
                width = self.rowWidth - left;
            end
            heading:SetLeft(left);
            heading:SetWidth(width);
            -- optimization to do: only need to do the following loop if heading left/width changed
            for row = 1, self.rows:GetItemCount() do
                local cell = self.rows:GetItem(row).cells[col];
                cell:SetLeft(left);
                cell:SetWidth(width);
            end
            weightExpended = weightExpended + weight;
            left = left + width + self.style.horizontalCellSpacing;
        end
    end
    self:_PlaceEdgeMovers();
end

function TableControl:SetSize(width, height)
    self:SetWidth(width);
    self:SetHeight(height);
end

function TableControl:SetWidth(width)
    Turbine.UI.Control.SetWidth(self, width);
    self.rowWidth = width;
    if (self.vScrollBar:IsVisible()) then
        self.rowWidth = width - 10;
        self.vScrollBar:SetLeft(self.rowWidth);
    end
    self.rows:SetWidth(self.rowWidth);
    for row = 1, self.rows:GetItemCount() do
        self.rows:GetItem(row):SetWidth(self.rowWidth);
    end
    self:_UpdateColumnWidths();
end

function TableControl:SetHeight(height)
    Turbine.UI.Control.SetHeight(self, height);
    self.rows:SetHeight(height - self.rows:GetTop());
    self.vScrollBar:SetHeight(height);
    self:_PlaceEdgeMovers(); -- update EdgeMover heights
end

function TableControl:_PlaceEdgeMovers()
    height = self:GetHeight();
    for e = 1, #self.edgeMovers do
        local edgeMover = self.edgeMovers[e];
        local headingToLeft = self.columns[edgeMover.leftColumn].heading;
        local headingToRight = self.columns[edgeMover.rightColumn].heading;

        local left = headingToLeft:GetLeft() + headingToLeft:GetWidth();
        local right = headingToRight:GetLeft();
        if (left > right) then -- columns overlap?
            left = right;
        end
        local center = (left + right) / 2;
        local width = edgeMover:GetWidth();
        left = math.floor(0.5 + center - (width / 2));

        edgeMover:SetLeft(left);
        edgeMover:SetHeight(height);
    end
end

-- For iterating over an array of cells in a for loop.  Example:
-- To change the color of every cell in the first two columns:
--     for cell in tblCtl:CellIterator(1, 1, nil, 2) do
--         cell:SetBackColor(Turbine.UI.Color.White);
--     end
function TableControl:CellIterator(firstRow, firstCol, lastRow, lastCol)
    if (firstRow == nil) then
        firstRow = 1;
    end
    if (lastRow == nil) then
        lastRow = self.rows:GetItemCount();
    end
    if (firstCol == nil) then
        firstCol = 1;
    end
    if (lastCol == nil) then
        lastCol = #self.columns;
    end

    local state = {
        ["self"] = self;
        ["firstRow"] = firstRow;
        ["firstCol"] = firstCol;
        ["lastRow"] = lastRow;
        ["lastCol"] = lastCol;
        ["row"] = firstRow;
        ["column"] = firstCol
    };
    local function iterator(state)
        if (state.row > state.lastRow) then
            return nil;
        end
        local cell = TableControl.GetCell(state.self, state.row, state.column);
        state.column = state.column + 1;
        if (state.column > state.lastCol) then
            state.column = state.firstCol;
            state.row = state.row + 1;
        end
        return cell;
    end
    return iterator, state, nil;
end

-- For calling member functions of many cells at once.  Examples:
--   minCellHeight = math.min(t:CellAggregator(1, 1, 1, 3):GetHeight());
--   t:CellAggregator():SetBackColor(Turbine.UI.Color(1, 1, 1, 1));
function TableControl:CellAggregator(firstRow, firstCol, lastRow, lastCol)
    local temp = {};
    temp.iterator = {self:CellIterator(firstRow, firstCol, lastRow, lastCol)};
    local temp_metatable = {};
    temp_metatable.__index = function(tab, key)
        rawset(tab, "funcName", key);
        return tab;
    end;
    temp_metatable.__call = function(tab, _, ...)
        local funcName = rawget(tab, "funcName");
        local results = {};
        for cell in unpack(rawget(tab, "iterator")) do
            local r = {cell[funcName](cell, ...)};
            for v in values(r) do
                table.insert(results, v);
            end
        end
        return unpack(results);
    end;
    setmetatable(temp, temp_metatable);
    return temp;
end

Thurallor = Thurallor or {};
Thurallor.UI = Thurallor.UI or {};
Thurallor.UI.TableControl = TableControl;
