ExpireTimeDialog = class(Turbine.UI.Lotro.Window);

function ExpireTimeDialog:Constructor(win, event, expTime)
    Turbine.UI.Lotro.Window.Constructor(self);
    self:SetText("Expiration Time");
    self:SetVisible(true);
    self:SetZOrder(win:GetZOrder() + 1);
    win:WaitForChild(self);

    local dateInfo = Turbine.Engine.GetDate();
    local dayOfWeek = dateInfo.DayOfWeek;
    self.weekdays = {};
    local dayNames = { "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" };
    for day = dayOfWeek, dayOfWeek + 7 do
        local text = "";
        if (day == dayOfWeek) then
            text = "Today (";
        elseif (day == dayOfWeek + 1) then
            text = "Tomorrow (";
        elseif (day == dayOfWeek + 7) then
            text = "Next ";
        end
        local dayNum = (day - 1) % 7 + 1;
        text = text .. dayNames[dayNum];
        if ((day == dayOfWeek) or (day == dayOfWeek + 1)) then
            text = text .. ")";
        end
        table.insert(self.weekdays, text);
    end
    self.monthNames = { "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December" };
    
    local font = Turbine.UI.Lotro.Font.Verdana14;
    local xmargin = 28;
    local parent = self;
    
    local function AddRadioButton(left, top, width, text, selected)
        local button = Thurallor.UI.RadioButton(parent, text, selected);
        button:SetFont(font);
        button:SetPosition(left, top);
        button:SetSize(width, 14);
        return button;
    end
    
    local function AddTextBox(left, top, width, value, minValue, maxValue)
        local textBox = Nudger(parent, font, value, minValue, maxValue);
        textBox:SetPosition(left, top);
        textBox:SetWidth(width);
        return textBox;
    end

    local function AddLabel(left, top, width, text)
        local label = Turbine.UI.Label();
        label:SetParent(parent);
        label:SetPosition(left, top + 11);
        label:SetFont(font);
        label:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
        label:SetText(text);
        label:SetSize(width, 24);
        return label;
    end
    
    local function AddPulldown(left, top, width, options, value)
        local dropDown = Thurallor.UI.DropDown(options, value);
        dropDown:SetParent(parent);
        dropDown:SetFont(font);
        dropDown:SetZOrder(1);
        dropDown:SetPosition(left, top + 14);
        dropDown:SetWidth(width);
        return dropDown;
    end

    local function AddButton(left, top, width, text)
        local button = Turbine.UI.Lotro.Button();
        button:SetParent(parent);
        button:SetText(text);
        button:SetPosition(left, top);
        button:SetWidth(width);
        return button;
    end

    -------------------------------
    -- * Delay from now:
    -------------------------------
    local left, top = xmargin, 44;
    self.delayFromNow = AddRadioButton(left, top, 200, " Delay from now:", false);
    self.delayFromNow.Select = function() self.delayFromNow:MouseClick(); end
    top = top + 20;
    left = xmargin + 16;
    self.delayFromNow.days = AddTextBox(left, top, 40, 0, 0, nil);
    self.delayFromNow.days.FocusGained = self.delayFromNow.Select;
    left = left + 40;
    AddLabel(left, top, 45, " days");
    left = left + 45;
    self.delayFromNow.hours = AddTextBox(left, top, 40, 0, 0, 23);
    self.delayFromNow.hours.FocusGained = self.delayFromNow.Select;
    left = left + 40;
    AddLabel(left, top, 51, " hours");
    left = left + 51;
    self.delayFromNow.minutes = AddTextBox(left, top, 40, 0, 0, 59);
    self.delayFromNow.minutes.FocusGained = self.delayFromNow.Select;
    left = left + 40;
    AddLabel(left, top, 64, " minutes");
    left = left + 64;
    self.delayFromNow.seconds = AddTextBox(left, top, 40, 0, 0, 59);
    self.delayFromNow.seconds.FocusGained = self.delayFromNow.Select;
    left = left + 40;
    AddLabel(left, top, 70, " seconds");
    local width = left + 70 + xmargin;
    top = top + 60;
    
    -------------------------------
    -- * Specific time this week:
    -------------------------------
    left = xmargin;
    self.timeThisWeek = AddRadioButton(left, top, 200, " Specific time this week:", true);
    self.timeThisWeek.Select = function() self.timeThisWeek:MouseClick(); end
    self.timeThisWeek.Clicked = function() self:ShowTimeThisWeek(); end
    top = top + 20;
    left = xmargin + 16;

    -- Make a control for grouping the following controls to allow easy show/hide of them all
    self.timeThisWeek.group = Turbine.UI.Control();
    self.timeThisWeek.group:SetParent(self);
    self.timeThisWeek.group:SetPosition(left, top);
    self.timeThisWeek.group:SetSize(width - 2 * xmargin - 16, 46);
    parent = self.timeThisWeek.group;
    top = 0; left = 0;

    self.selectedDayOfWeek = 1;
    self.timeThisWeek.dayOfWeek = AddPulldown(left, top, 188, self.weekdays, self.weekdays[1]);
    AddCallback(self.timeThisWeek.dayOfWeek, "MouseClick", function()
        self.timeThisWeek.Select();
        self:Focus();
    end);
    self.timeThisWeek.dayOfWeek.ItemChanged = function(_, args)
        self.selectedDayOfWeek = args.Index;
    end
    left = left + 188;
    AddLabel(left, top, 25, " at");
    left = left + 25;
    self.timeThisWeek.hour = AddTextBox(left, top, 30, 0, 0, 23);
    self.timeThisWeek.hour.FocusGained = self.timeThisWeek.Select;
    left = left + 30;
    AddLabel(left, top, 6, ":");
    left = left + 6;
    self.timeThisWeek.minute = AddTextBox(left, top, 30, 0, 0, 59);
    self.timeThisWeek.minute.FocusGained = self.timeThisWeek.Select;
    left = left + 30;
    AddLabel(left, top, 70, " o'clock");
    self.timeThisWeek.second = Turbine.UI.TextBox(); -- hidden
    parent = self;
    top = self.timeThisWeek.group:GetTop() + 60;

    -------------------------------
    -- * Specific date and time:
    -------------------------------
    left = xmargin;
    self.dateAndTime = AddRadioButton(left, top, 200, " Specific date and time:", false);
    self.dateAndTime.Select = function() self.dateAndTime:MouseClick(); end
    top = top + 20;
    left = xmargin + 16;
    self.dateAndTime.day = AddTextBox(left, top, 30, 1, 1, 31);
    self.dateAndTime.day.FocusGained = self.dateAndTime.Select;
    left = left + 34;
    self.selectedMonth = 1;
    self.dateAndTime.month = AddPulldown(left, top, 100, self.monthNames, self.monthNames[1]);
    AddCallback(self.dateAndTime.month, "MouseClick", function()
        self.dateAndTime.Select();
        self:Focus();
    end);
    self.dateAndTime.month.ItemChanged = function(_, args)
        self.selectedMonth = args.Index;
    end
    left = left + 104;
    self.dateAndTime.year = AddTextBox(left, top, 50, 2017, 2017, nil);
    self.dateAndTime.year.FocusGained = self.dateAndTime.Select;
    left = left + 50;
    AddLabel(left, top, 25, " at");
    left = left + 25;
    self.dateAndTime.hour = AddTextBox(left, top, 30, 0, 0, 23);
    self.dateAndTime.hour.FocusGained = self.dateAndTime.Select;
    left = left + 30;
    AddLabel(left, top, 6, ":");
    left = left + 6;
    self.dateAndTime.minute = AddTextBox(left, top, 30, 0, 0, 59);
    self.dateAndTime.minute.FocusGained = self.dateAndTime.Select;
    left = left + 30;
    AddLabel(left, top, 70, " o'clock");
    self.dateAndTime.second = Turbine.UI.TextBox(); -- hidden
    top = top + 64;
    
    Thurallor.UI.RadioButton.LinkPeers({ self.delayFromNow, self.timeThisWeek, self.dateAndTime });

    left = math.floor(0.5 + (width - 166) / 2);
    self.okButton = AddButton(left, top, 80, "Ok");
    self.okButton.SetEnabled = function(_, enable)
        -- Lotro.Button can't get mouse events while disabled.  Need to superimpose another control.
        if (enable and not self.okButton:IsEnabled()) then
            self.okMask:SetParent(nil);
            self.okMask = nil;
        elseif (not enable and self.okButton:IsEnabled()) then
            self.okMask = Turbine.UI.Control();
            self.okMask:SetParent(self.okButton);
            self.okMask:SetSize(self.okButton:GetSize());
            Thurallor.UI.Tooltip("The selected time has already passed."):Attach(self.okMask);
        end
        Turbine.UI.Lotro.Button.SetEnabled(self.okButton, enable);
    end
    self.okButton.Click = function()
        self:Update();
        local method = "delayFromNow";
        if (self.timeThisWeek:IsChecked()) then
            method = "timeThisWeek";
        elseif (self.dateAndTime:IsChecked()) then
            method = "dateAndTime";
        end
        DoCallbacks(self, "Ok", { expTime = self.expTime; method = method });
        self:Close();
    end
    left = left + 86;
    
    self.cancelButton = AddButton(left, top, 80, "Cancel");
    self.cancelButton.Click = function()
        self:Close();
    end
    top = top + self.cancelButton:GetHeight();

    local height = top + 22;
    self:SetSize(width, height);
    local x, y = Turbine.UI.Display.GetMousePosition();
    local left = math.max(0, x - (width / 2));
    local top = math.max(0, y - (height / 2));
    self:SetPosition(left, top);

    if not (self:SetTimeThisWeek(expTime)) then
        self:SetDateAndTime(expTime);
        self.timeThisWeek:SetChecked(false);
        self.dateAndTime:SetChecked(true);
    end
    self:SetWantsUpdates(true);
    self:Update();
end

function ExpireTimeDialog:Update()
    local gameTime = Turbine.Engine.GetGameTime();
    if (self.delayFromNow:IsChecked()) then
        expTime = self:ReadDelayFromNow();
    elseif (self.timeThisWeek:IsChecked()) then
        expTime = self:ReadTimeThisWeek();
    else -- self.dateAndTime:IsChecked()
        expTime = self:ReadDateAndTime();
    end
    if (expTime < gameTime) then
        self.okButton:SetEnabled(false);
        expTime = gameTime;
    else
        self.okButton:SetEnabled(true);
    end
    if (self.delayFromNow:IsChecked()) then
        self:SetTimeThisWeek(expTime);
        self:SetDateAndTime(expTime);
    elseif (self.timeThisWeek:IsChecked()) then
        self:SetDelayFromNow(expTime);
        self:SetDateAndTime(expTime);
    else -- self.dateAndTime:IsChecked()
        self:SetDelayFromNow(expTime);
        self:SetTimeThisWeek(expTime);
    end
end

function ExpireTimeDialog:ReadDelayFromNow()
    local expTime = Turbine.Engine.GetGameTime();
    expTime = expTime + tonumber(self.delayFromNow.days:GetText()) * 24 * 60 * 60;
    expTime = expTime + tonumber(self.delayFromNow.hours:GetText()) * 60 * 60;
    expTime = expTime + tonumber(self.delayFromNow.minutes:GetText()) * 60;
    expTime = expTime + tonumber(self.delayFromNow.seconds:GetText());
    return expTime;
end

function ExpireTimeDialog:SetDelayFromNow(expTime)
    self.expTime = expTime;
    remaining = math.ceil(expTime - Turbine.Engine.GetGameTime());
    local s = math.floor(remaining % 60);
    local m = math.floor((remaining % 3600) / 60);
    local h = math.floor((remaining % 86400) / 3600);
    local d = math.floor(remaining / 86400);
    self.delayFromNow.days:SetText(d);
    self.delayFromNow.hours:SetText(h);
    self.delayFromNow.minutes:SetText(m);
    self.delayFromNow.seconds:SetText(s);
end

function ExpireTimeDialog:ReadTimeThisWeek()
    local expTime = Turbine.Engine.GetGameTime();
    local dateInfo = Turbine.Engine.GetDate();
    local daysAdvanced = self.selectedDayOfWeek - 1;
    local hoursAdvanced = tonumber(self.timeThisWeek.hour:GetText()) - dateInfo["Hour"];
    local minutesAdvanced = tonumber(self.timeThisWeek.minute:GetText()) - dateInfo["Minute"];
    local secondsAdvanced = tonumber(self.timeThisWeek.second:GetText()) - dateInfo["Second"];
    expTime = expTime + daysAdvanced * 24* 60 * 60;
    expTime = expTime + hoursAdvanced * 60 * 60;
    expTime = expTime + minutesAdvanced * 60;
    expTime = expTime + secondsAdvanced;
    return expTime;
end

function ExpireTimeDialog:SetTimeThisWeek(expTime)
    self.expTime = expTime;
    remaining = math.ceil(expTime - Turbine.Engine.GetGameTime());
    local s = math.floor(remaining % 60);
    local m = math.floor((remaining % 3600) / 60);
    local h = math.floor((remaining % 86400) / 3600);
    local d = math.floor(remaining / 86400);
    local dateInfo = Turbine.Engine.GetDate();
    local day = d;
    local hour = h + dateInfo.Hour;
    local minute = m + dateInfo.Minute;
    local second = s + dateInfo.Second;
    if (second >= 60) then
        minute = minute + 1;
        second = second - 60;
    end
    if (minute >= 60) then
        hour = hour + 1;
        minute = minute - 60;
    end
    if (hour >= 24) then
        day = day + 1;
        hour = hour - 24;
    end
    self.selectedDayOfWeek = day + 1;
    local weekday = self.weekdays[day + 1];
    if (weekday) then
        self:ShowTimeThisWeek();
        if (self.timeThisWeek.dayOfWeek:GetText() ~= weekday) then
            self.timeThisWeek.dayOfWeek:SetText(self.weekdays[day + 1]);
        end
        self.timeThisWeek.hour:SetText(hour);
        if (minute < 10) then
            minute = "0" .. minute;
        end
        self.timeThisWeek.minute:SetText(minute);
        self.timeThisWeek.second:SetText(second);
        return true;
    else
        self:HideTimeThisWeek();
        return false;
    end
end

function ExpireTimeDialog:HideTimeThisWeek()
    if (self.timeThisWeek.group:GetParent() ~= nil) then
        self.timeThisWeek.group:SetParent(nil);
        self.timeThisWeek:SetText(" Specific time this week");
    end
end

function ExpireTimeDialog:ShowTimeThisWeek()
    if (self.timeThisWeek.group:GetParent() == nil) then
        self.timeThisWeek.group:SetParent(self);
        self.timeThisWeek:SetText(" Specific time this week:");
    end
end

function ExpireTimeDialog:ReadDateAndTime()
    local dateInfo = Turbine.Engine.GetDate();
    local dayOfMonth = tonumber(self.dateAndTime.day:GetText());
    local month = self.selectedMonth;
    local year = tonumber(self.dateAndTime.year:GetText());
    local d = self:GetDayOfYear(year, month, dayOfMonth);
    d = d - self:GetDayOfYear(dateInfo.Year, dateInfo.Month, dateInfo.Day);
    while true do
        local yearLength = 365;
        if (self:IsLeapYear(dateInfo.Year)) then
            yearLength = 366;
        end
        if (dateInfo.Year < year) then
            d = d + yearLength;
            dateInfo.Year = dateInfo.Year + 1;
        else
            break;
        end
    end
    local h = tonumber(self.dateAndTime.hour:GetText()) - dateInfo.Hour;
    local m = tonumber(self.dateAndTime.minute:GetText()) - dateInfo.Minute;
    local s = tonumber(self.dateAndTime.second:GetText()) - dateInfo.Second;
    local remaining = (d * 24 * 60 * 60) + (h * 60 * 60) + (m * 60) + s;
    local expTime = Turbine.Engine.GetGameTime() + remaining;
    return expTime;
end

function ExpireTimeDialog:SetDateAndTime(expTime)
    self.expTime = expTime;
    remaining = math.ceil(expTime - Turbine.Engine.GetGameTime());
    local s = math.floor(remaining % 60);
    local m = math.floor((remaining % 3600) / 60);
    local h = math.floor((remaining % 86400) / 3600);
    local d = math.floor(remaining / 86400);
    local dateInfo = Turbine.Engine.GetDate();
    local year = dateInfo.Year;
    local day = d + dateInfo.DayOfYear;
    local hour = h + dateInfo.Hour;
    local minute = m + dateInfo.Minute;
    local second = s + dateInfo.Second;
    if (second >= 60) then
        minute = minute + 1;
        second = second - 60;
    end
    if (minute >= 60) then
        hour = hour + 1;
        minute = minute - 60;
    end
    if (hour >= 24) then
        day = day + 1;
        hour = hour - 24;
    end
    while true do
        local yearLength = 365;
        if (self:IsLeapYear(year)) then
            yearLength = 366;
        end
        if (day >= yearLength) then
            year = year + 1;
            day = day - yearLength;
        else
            break;
        end
    end
    local month, dayOfMonth = self:GetMonthAndDay(year, day);
    local monthName = self.monthNames[month];
    self.selectedMonth = month;
    self.dateAndTime.day:SetText(dayOfMonth);
    if (self.dateAndTime.month:GetText() ~= monthName) then
        self.dateAndTime.month:SetText(monthName);
    end
    self.dateAndTime.year:SetText(year);
    self.dateAndTime.hour:SetText(hour);
    if (minute < 10) then
        minute = "0" .. minute;
    end
    self.dateAndTime.second:SetText(second);
    self.dateAndTime.minute:SetText(minute);
end

function ExpireTimeDialog:IsLeapYear(year)
    local div4 = ((year % 4) == 0);
    local div100 = ((year % 100) == 0);
    local div400 = ((year % 400) == 0);
    if (div4) then
        if (div100) then
            return div400;
        else
            return true;
        end
    end
end

function ExpireTimeDialog:GetMonthAndDay(year, dayOfYear)
    local monthLengths = {
        ["January"] = 31;
        ["February"] = 28;
        ["March"] = 31;
        ["April"] = 30;
        ["May"] = 31;
        ["June"] = 30;
        ["July"] = 31;
        ["August"] = 31;
        ["September"] = 30;
        ["October"] = 31;
        ["November"] = 30;
        ["December"] = 31;
    };
    if (self:IsLeapYear(year)) then
        monthLengths.February = 29;
    end
    local month, dayOfMonth = 1, dayOfYear;
    while true do
        local monthLength = monthLengths[self.monthNames[month]];
        if (dayOfMonth > monthLength) then
            month = month + 1;
            dayOfMonth = dayOfMonth - monthLength;
        else
            break;
        end
    end
    return month, dayOfMonth;
end

function ExpireTimeDialog:GetDayOfYear(year, month, dayOfMonth)
    local monthLengths = {
        ["January"] = 31;
        ["February"] = 28;
        ["March"] = 31;
        ["April"] = 30;
        ["May"] = 31;
        ["June"] = 30;
        ["July"] = 31;
        ["August"] = 31;
        ["September"] = 30;
        ["October"] = 31;
        ["November"] = 30;
        ["December"] = 31;
    };
    if (self:IsLeapYear(year)) then
        monthLengths.February = 29;
    end
    local dayOfYear = dayOfMonth;
    while (month > 1) do
        dayOfYear = dayOfYear + monthLengths[self.monthNames[month - 1]];
        month = month - 1;
    end
    return dayOfYear;
end
