OptionsPanel = class(Turbine.UI.Control);

local importPath = getfenv(1)._.Name;
local imagePath = string.gsub(string.gsub(importPath, "%.OptionsPanel$", ""), "%.", "/") .. "/Images/";

function OptionsPanel:Constructor(window, settings)
    Turbine.UI.Control.Constructor(self);
    self.window = window;
    self.settings = settings;

    self.player = Turbine.Gameplay.LocalPlayer.GetInstance();
    self.selectedChar = self.player:GetName();
    
    local instruct = Turbine.UI.Label();
    instruct:SetParent(self);
    instruct:SetFont(Turbine.UI.Lotro.Font.Verdana14);
    instruct:SetForeColor(Turbine.UI.Color.PaleGoldenrod);
    instruct:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
    instruct:SetText("These are the options for the Reminders plugin:");
    instruct:SetSize(600, 24);
    self.instruct = label;

    self.tabStack = Thurallor.UI.TabCardStack();
    self.tabStack:SetParent(self);
    self.tabStack:SetPosition(4, 24);
    
    local globalTab = Thurallor.UI.TabCard(false, false, false);
    globalTab:SetTabText("Global");
    self.globalPanel = Turbine.UI.Control();
    globalTab:SetInteriorControl(self.globalPanel);
    globalTab:SetInteriorAlignment(Turbine.UI.ContentAlignment.TopLeft);
    self.tabStack:AddTabCard(globalTab);
    self.globalTab = globalTab;

    local charTab = Thurallor.UI.TabCard(false, false, false);
    self.charPanel = Turbine.UI.Control();
    charTab:SetInteriorControl(self.charPanel);
    charTab:SetInteriorAlignment(Turbine.UI.ContentAlignment.TopLeft);
    charTab:SetTabLeft(1);
    charTab:SetTabText("Character");
    self.tabStack:AddTabCard(charTab);
    self.charTab = charTab;

    AddCallback(globalTab, "Click", function(_, args)
        if (args.Button == Turbine.UI.MouseButton.Left) then
            self:UpdateDisplay();
        end
    end);
    AddCallback(charTab, "Click", function(_, args)
        if (args.Button == Turbine.UI.MouseButton.Left) then
            self:UpdateDisplay();
        end
    end);
    
    globalTab:BringToFront();
end

function OptionsPanel:Show()
    Turbine.PluginManager.ShowOptions(Plugins.Reminders);
    self:SetWantsUpdates(); -- start pulsating the background
end

function OptionsPanel:MouseEnter()
    self:SetWantsUpdates(false); -- stop pulsating the background when mouse arrives
    self:SetBackColor(nil);
end

function OptionsPanel:Update()
    -- Pulsate the background to draw the user's eye
    local gameTime = Turbine.Engine.GetGameTime();
    local _, progress = math.modf(gameTime);
    if (progress < 0.5) then
        progress = 0.5 - progress;
    else
        progress = progress - 0.5;
    end
    self:SetBackColor(Turbine.UI.Color(1, progress, progress, progress));
end

function OptionsPanel:SaveSettings()
    self.window:SaveSettings();
end

function OptionsPanel:SizeChanged()
    local width = self:GetWidth();
    self.tabStack:SetWidth(width - 7);
    self.tabStack:RedistributeTabs();
    self.globalPanel:SetWidth(width - 12);
    self.charPanel:SetWidth(width - 12);
end

function OptionsPanel:UpdateDisplay()
    if (self.colorPicker) then
        self.colorPicker:Close();
        self.colorPicker = nil;
    end
    if (self.globalTab:IsInFront()) then
        self:UpdateGlobalDisplay();
    else -- (self.charTab:IsInFront())
        self:UpdateCharDisplay();
    end
    local height = math.max(self.globalPanel:GetHeight(), self.charPanel:GetHeight());
    self.tabStack:SetHeight(height + 30);
    self:SetHeight(self.tabStack:GetTop() + height + 40);
end

function OptionsPanel:UpdateCharDisplay()
    local panel = self.charPanel;
    local top = 16;
    local heading, checkbox, label, swatch, button, dropDown;
    local charSettings = self.settings.characters[self.selectedChar];
    local tabNames, tabExists = self.window:GetUserTabs();
    local catNames, catExists = self.window:GetUserCategories();
    local tabName, catName;
    
    -- Remove child controls and SizeChanged callbacks; we'll recreate them now
    panel:GetControls():Clear();
    panel.SizeChanged = nil;

    dropDown, label, top = self:AddDropDown(panel, top, 200, self.window.characters, self.selectedChar, "Editing options for");
    dropDown.SelectionChanged = function(_, args)
        self.selectedChar = args.Item;
        self:UpdateDisplay();
    end

    heading, top = self:AddHeading(panel, top, "Standard Automatic Reminders");
    
    local function CheckAutoReminderParams(name)
        if (not charSettings.autoReminders[name]) then
            charSettings.autoReminders[name] = {};
        end
        tabName = charSettings.autoReminders[name].tab;
        if ((not tabName) or (not tabExists[tabName])) then
            charSettings.autoReminders[name].tab = tabNames[1];
        end
        catName = charSettings.autoReminders[name].category;
        if ((not catName) or (not catExists[catName])) then
            charSettings.autoReminders[name].category = catNames[1];
        end
    end

    CheckAutoReminderParams("crafting");
    checkbox, top = self:AddCheckBox(panel, top, "Create reminders for crafting cooldowns:", charSettings.autoReminders.crafting.enabled);
    checkbox.CheckedChanged = function(ctl)
        charSettings.autoReminders.crafting.enabled = ctl:IsChecked();
        self:SaveSettings();
    end

    dropDown, label, top = self:AddDropDown(panel, top, 200, tabNames, charSettings.autoReminders.crafting.tab, "under Tab");
    dropDown.SelectionChanged = function(ctl, args)
        charSettings.autoReminders.crafting.tab = args.Item;
        self:SaveSettings();
    end

    dropDown, label, top = self:AddDropDown(panel, top, 200, catNames, charSettings.autoReminders.crafting.category, "in Category");
    dropDown.SelectionChanged = function(ctl, args)
        charSettings.autoReminders.crafting.category = args.Item;
        self:SaveSettings();
    end

    CheckAutoReminderParams("crates");
    checkbox, top = self:AddCheckBox(panel, top, "Create reminders for ingredient crates:", charSettings.autoReminders.crates.enabled);
    checkbox.CheckedChanged = function(ctl)
        charSettings.autoReminders.crates.enabled = ctl:IsChecked();
        self:SaveSettings();
    end

    dropDown, label, top = self:AddDropDown(panel, top, 200, tabNames, charSettings.autoReminders.crates.tab, "under Tab");
    dropDown.SelectionChanged = function(ctl, args)
        charSettings.autoReminders.crates.tab = args.Item;
        self:SaveSettings();
    end

    dropDown, label, top = self:AddDropDown(panel, top, 200, catNames, charSettings.autoReminders.crates.category, "in Category");
    dropDown.SelectionChanged = function(ctl, args)
        charSettings.autoReminders.crates.category = args.Item;
        self:SaveSettings();
    end
    
    panel:SetHeight(top);
    DoCallbacks(panel, "SizeChanged");
end

function OptionsPanel:UpdateGlobalDisplay()
    local panel = self.globalPanel;
    local top = 0;
    local heading, checkbox, label, swatch, button, dropDown;
    local catNames, catExists = self.window:GetUserCategories();

    -- Remove child controls and SizeChanged callbacks; we'll recreate them now
    panel:GetControls():Clear();
    panel.SizeChanged = nil;

    local heading, top = self:AddHeading(panel, top, "General");
    
    checkbox, top = self:AddCheckBox(panel, top, "Pause the plugin during combat", self.settings.pauseDuringCombat);
    checkbox.CheckedChanged = function()
        self.settings.pauseDuringCombat = checkbox:IsChecked();
        self.window:InCombatChanged();
        self:SaveSettings();
    end

    top = top + 10; -- skip some space
    label, top = self:AddLabel(panel, top, "When creating new reminders:");
    
    local defaultCat = self.settings.defaultCategory;
    if ((not defaultCat) or (not catExists[defaultCat])) then
        self.settings.defaultCategory = catNames[1];
    end
    dropDown, label, top = self:AddDropDown(panel, top, 200, catNames, self.settings.defaultCategory, "Default category");
    dropDown.SelectionChanged = function(_, args)
        self.settings.defaultCategory = args.Item;
    end

    self.settings.defaultDelay = self.settings.defaultDelay or 0;
    textBox, label, label, top = self:AddTextBox(panel, top, 40, "Default time", self.settings.defaultDelay, "hours");
    textBox.FocusLost = function(ctl)
        local value = tonumber(ctl:GetText());
        self.settings.defaultDelay = value or 0;
        if (value == nil) then
            self:UpdateDisplay();
        end
        self:SaveSettings();
    end

    top = top + 10; -- skip some space
    label, top = self:AddLabel(panel, top, "When postponing expired reminders:");
    
    self.settings.postponeDelay = self.settings.postponeDelay or 4;
    textBox, label, label, top = self:AddTextBox(panel, top, 40, "Default time", self.settings.postponeDelay, "hours");
    textBox.FocusLost = function(ctl)
        local value = tonumber(ctl:GetText());
        self.settings.postponeDelay = value or 0;
        if (value == nil) then
            self:UpdateDisplay();
        end
        self:SaveSettings();
    end
    
    heading, top = self:AddHeading(panel, top, "Categories");
    
    local function AddCategory(top, text, color)
        color = Turbine.UI.Color(unpack(color));
    
        local button = Turbine.UI.Control();
        button:SetParent(panel);
        button:SetSize(16, 16);
        button:SetTop(top + 4);
        button:SetBackground(0x41000196);
        button.MouseEnter = function() button:SetBackground(0x41000198) end;
        button.MouseLeave = function() button:SetBackground(0x41000196) end;
        button.MouseDown = function() button:SetBackground(0x41000197) end;
        button.MouseUp = function() button:SetBackground(0x41000198) end;
        Thurallor.UI.Tooltip("Delete this category"):Attach(button);
        
        local swatch = Turbine.UI.Control();
        swatch:SetParent(panel);
        swatch:SetSize(16, 16);
        swatch:SetTop(top + 4);
        swatch:SetBackColor(Turbine.UI.Color.Gray);
        swatch.MouseEnter = function() swatch:SetBackColor(Turbine.UI.Color.White) end;
        swatch.MouseLeave = function() swatch:SetBackColor(Turbine.UI.Color.Gray) end;
        Thurallor.UI.Tooltip("Change the color for this category"):Attach(swatch);
        
        local inside = Turbine.UI.Control();
        inside:SetParent(swatch);
        inside:SetMouseVisible(false);
        inside:SetBackColor(color);
        inside:SetPosition(1, 1);
        inside:SetSize(14, 14);
        swatch.inside = inside;
        
        local label = Turbine.UI.Lotro.TextBox();
        label:SetParent(panel);
        label:SetSelectable(true);
        label:SetMultiline(false);
        label:SetFont(Turbine.UI.Lotro.Font.Verdana14);
        label:SetForeColor(color);
        label:SetBackColor(Turbine.UI.Color(1, 0, 0.19, 0.33));
        label:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
        label:SetText(text);
        label:SetPosition(0, top);
        label:SetSize(200, 24);
        label.MouseEnter = function() label:SetBackColor(Turbine.UI.Color(1, 0, 0.57, 1)) end;
        label.MouseLeave = function() label:SetBackColor(Turbine.UI.Color(1, 0, 0.19, 0.33)) end;
        
        AddCallback(panel, "SizeChanged", function()
            local width = panel:GetWidth();
            local center = math.floor(0.5 + width / 2);
            swatch:SetLeft(math.max(0, center - 150));
            label:SetLeft(swatch:GetLeft() + 20);
            label:SetWidth(math.min(264, width - label:GetLeft() - 20));
            button:SetLeft(label:GetLeft() + label:GetWidth() + 4);
        end);
        return label, swatch, button, top + 23;
    end
    
    for _, name in ipairs(catNames) do
        local n = catExists[name];
        local color = self.settings.categories[n].color;
        label, swatch, button, top = AddCategory(top, name, color);
        swatch.n, swatch.label = n, label;
        swatch.MouseClick = function(ctl, args)
            if (args.Button == Turbine.UI.MouseButton.Left) then
                self:PickCategoryColor(ctl.n, ctl.inside, ctl.label);
            end
        end
        label.n = n;
        label.FocusLost = function(ctl)
            self:CategoryRenamed(ctl.n, ctl:GetText());
        end
        button.n, button.label = n, label;
        button.MouseClick = function(ctl, args)
            if (args.Button == Turbine.UI.MouseButton.Left) then
                ctl.label.FocusLost = nil;
                self:CategoryDeleted(ctl.n);
            end
        end
        -- Don't let the user delete the last category
        if (#catNames == 1) then
            button:SetParent(nil);
        end
    end
    top = top + 5;
    
    button, top = self:AddButton(panel, top, 120, "New Category");
    button.Click = function()
        self:NewCategory();
    end
    
    panel:SetHeight(top);
    DoCallbacks(panel, "SizeChanged");
end

function OptionsPanel:AddTextBox(panel, top, width, leftDesc, value, rightDesc)
    local leftLabel = Turbine.UI.Label();
    leftLabel:SetParent(panel);
    leftLabel:SetFont(Turbine.UI.Lotro.Font.Verdana14);
    leftLabel:SetForeColor(Turbine.UI.Color.PaleGoldenrod);
    leftLabel:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleRight);
    leftLabel:SetPosition(0, top);
    leftLabel:SetHeight(22);
    leftLabel:SetMultiline(false);
    leftLabel:SetText(leftDesc);
    local textBox = Turbine.UI.Lotro.TextBox();
    textBox:SetParent(panel);
    textBox:SetSelectable(true);
    textBox:SetMultiline(false);
    textBox:SetFont(Turbine.UI.Lotro.Font.Verdana14);
    textBox:SetForeColor(Turbine.UI.Color.PaleGoldenrod);
    textBox:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
    textBox:SetText(value);
    textBox:SetPosition(0, top + 1);
    textBox:SetSize(width, 21);
    local rightLabel = Turbine.UI.Label();
    rightLabel:SetParent(panel);
    rightLabel:SetFont(Turbine.UI.Lotro.Font.Verdana14);
    rightLabel:SetForeColor(Turbine.UI.Color.PaleGoldenrod);
    rightLabel:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
    rightLabel:SetTop(top);
    rightLabel:SetHeight(22);
    rightLabel:SetMultiline(false);
    rightLabel:SetText(rightDesc);
    AddCallback(panel, "SizeChanged", function()
        local center = math.floor(0.5 + panel:GetWidth() / 2);
        leftLabel:SetWidth(center - 2);
        textBox:SetLeft(center + 2);
        rightLabel:SetLeft(center + 2 + textBox:GetWidth() + 4);
        rightLabel:SetWidth(center); -- goes off the edge; don't care
    end);
    return textBox, label, label, top + 24;
end

function OptionsPanel:AddDropDown(panel, top, width, items, selectedItem, desc)
    local label = Turbine.UI.Label();
    label:SetParent(panel);
    label:SetFont(Turbine.UI.Lotro.Font.Verdana14);
    label:SetForeColor(Turbine.UI.Color.PaleGoldenrod);
    label:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleRight);
    label:SetPosition(0, top);
    label:SetHeight(22);
    label:SetMultiline(false);
    label:SetText(desc);
    local dropDown = Thurallor.UI.DropDown(items, selectedItem);
    dropDown:SetParent(panel);
    dropDown:SetFont(Turbine.UI.Lotro.Font.Verdana14);
    dropDown:SetTop(top);
    dropDown:SetSize(width, 22);
    dropDown:SetExpandedWidth(width);
    dropDown:SetAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
    AddCallback(panel, "SizeChanged", function()
        local center = math.floor(0.5 + panel:GetWidth() / 2);
        label:SetWidth(center - 2);
        dropDown:SetLeft(center + 2);
        dropDown:SetWidth(math.min(center - 6, width));
    end);
    return dropDown, label, top + 24;
end

function OptionsPanel:AddHeading(panel, top, text)
    -- Thanks Pengoros for this pretty heading style :-)
    local label = Turbine.UI.Label();
    label:SetParent(panel);
    label:SetSize(400, 30);
    label:SetFont(Turbine.UI.Lotro.Font.TrajanPro18);
    label:SetForeColor(Turbine.UI.Color(244/255, 255/255, 51/255));
    label:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
    label:SetFontStyle(Turbine.UI.FontStyle.Outline);
    label:SetOutlineColor(Turbine.UI.Color.Black);
    label:SetText(text);
    label:SetBlendMode(Turbine.UI.BlendMode.Overlay);
    label:SetBackground(imagePath .. "options_panel_divider.tga");
    label:SetTop(top + 15);
    AddCallback(panel, "SizeChanged", function()
        local center = math.floor(0.5 + panel:GetWidth() / 2);
        label:SetLeft(center - 200);
    end);
    return label, top + 56;
end

function OptionsPanel:AddCheckBox(panel, top, text, checked)
    local checkbox = Turbine.UI.Lotro.CheckBox();
    checkbox:SetParent(panel);
    checkbox:SetFont(Turbine.UI.Lotro.Font.Verdana14);
    checkbox:SetForeColor(Turbine.UI.Color.PaleGoldenrod);
    checkbox:SetText(" " .. text);
    checkbox:SetChecked(checked);
    checkbox:SetTop(top + 3);
    checkbox:SetHeight(16);
    AddCallback(panel, "SizeChanged", function()
        local center = math.floor(0.5 + panel:GetWidth() / 2);
        checkbox:SetLeft(math.max(0, center - 200));
        checkbox:SetWidth(center + 200);
    end);
    return checkbox, top + 24;
end

function OptionsPanel:AddLabel(panel, top, text)
    local label = Turbine.UI.Label();
    label:SetParent(panel);
    label:SetFont(Turbine.UI.Lotro.Font.Verdana14);
    label:SetForeColor(Turbine.UI.Color.PaleGoldenrod);
    label:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
    label:SetText(text);
    label:SetTop(top);
    label:SetHeight(24);
    AddCallback(panel, "SizeChanged", function()
        local center = math.floor(0.5 + panel:GetWidth() / 2);
        label:SetLeft(math.max(0, center - 200));
        label:SetWidth(center + 200);
    end);
    return checkbox, top + 24;
end

function OptionsPanel:AddButton(panel, top, width, text)
    local button = Turbine.UI.Lotro.Button();
    button:SetParent(panel);
    button:SetText(text);
    button:SetTop(top);
    button:SetWidth(width);
    AddCallback(panel, "SizeChanged", function()
        local center = math.floor(0.5 + panel:GetWidth() / 2);
        button:SetLeft(center - math.floor(button:GetWidth() / 2));
    end);
    return button, top + button:GetHeight();
end

function OptionsPanel:NewCategory()
    local catName, catExists = self.window:GetUserCategories();
    local nameStem = "New category";
    local num = 1;
    local name = nameStem;
    while catExists[name] do
        num = num + 1;
        name = nameStem .. " (" .. num .. ")";
    end
    local new = { name = name; color = { 1, 0.75, 0.75, 0.75 } };
    table.insert(self.settings.categories, new);
    self:SaveSettings();
    self:UpdateDisplay();
end

function OptionsPanel:CategoryDeleted(n)
    local name = self.settings.categories[n].name;
    table.remove(self.settings.categories, n);
    self:SaveSettings();
    self:UpdateDisplay();

    -- Find all existing events with this category and change them to the default
    self.window:ForEachEventDo(function(event)
        local cat = event:GetCategory();
        if (cat == name) then
            event:SetCategory(self.settings.defaultCategory);
        end
    end);
end

function OptionsPanel:CategoryRenamed(n, new)
    local old = self.settings.categories[n].name;
    if (old == new) then
        -- Nothing to do
        return;
    end
    -- Check for name collisions
    self.settings.categories[n].name = ""; -- invalid name, temporarily
    local catName, catExists = self.window:GetUserCategories();
    if (catExists[new]) then
        -- Name collision detected.  Restore old name.
        self.settings.categories[n].name = old;
    else
        -- New name is unique.
        self.settings.categories[n].name = new;
        
        -- Update default category for new reminders
        if (self.settings.defaultCategory == old) then
            self.settings.defaultCategory = new;
        end
        
        -- Update category for automatic reminders
        for charName, _ in pairs(self.settings.characters) do
            local charSettings = self.settings.characters[charName];
            for _, tbl in ipairs({
                charSettings.autoReminders.crafting
            }) do
                if (tbl.category == old) then
                    tbl.category = new;
                end
            end
        end
        
        -- Update category name in existing events
        self.window:ForEachEventDo(function(event)
            local cat = event:GetCategory();
            if (cat == old) then
                event:SetCategory(new);
            end
        end);
        self:SaveSettings();
    end
    self:UpdateDisplay();
end

function OptionsPanel:CategoryColorChanged(n, color)
    local name = self.settings.categories[n].name;
    self.settings.categories[n].color = { 1, color.R; color.G; color.B };
    self:SaveSettings();
    self:UpdateDisplay();

    -- Apply new color to all existing events
    self.window:ForEachEventDo(function(event)
        local cat = event:GetCategory();
        if (cat == name) then
            event:SetCategory(name);
        end
    end);
end

function OptionsPanel:PickCategoryColor(n, swatch, label)
    if (self.colorPicker) then
        self.colorPicker:Close();
    end
    self.colorPicker = Thurallor.UI.ColorPicker(swatch:GetBackColor());
    self.colorPicker.Accepted = function(picker)
        local color = picker:GetColor();
        self:CategoryColorChanged(n, color);
        self.colorPicker = nil;
    end
    self.colorPicker.ColorChanged = function(picker)
        local color = picker:GetColor();
        swatch:SetBackColor(color);
        label:SetForeColor(color);
    end
    self.colorPicker.Canceled = self.colorPicker.ColorChanged;
end

