Tab = class(Thurallor.UI.TabCard);

local importPath = getfenv(1)._.Name;
local imagePath = string.gsub(string.gsub(importPath, "%.Tab$", ""), "%.", "/") .. "/Images/";

function Tab:Constructor(window, settings)
    Thurallor.UI.TabCard.Constructor(self, false, false, true);
    self.window = window;

    AddCallback(self.tabText, "MouseClick", function(_, args)
        if (args.Button == Turbine.UI.MouseButton.Left) then
            self.window.settings.frontTab = self.settings.name;
            self:SaveSettings();
        elseif (args.Button == Turbine.UI.MouseButton.Right) then
            self:ShowTabContextMenu();
        end
    end);
    AddCallback(self.tabText, "MouseEnter", function(ctl)
        self.window.shimmering:Remove(ctl);
    end);

    -- Default settings
    local defaults = {};
    defaults.presentation = {
        columns = {
            { "desc", 3 }, { "category", 1 }, { "charName", 1 }, { "remaining", 1 }
        };
        sort = { column = "remaining"; dir = false };
    };
    defaults.tabLeft = 0;
    defaults.events = {};

    -- Previously-saved settings override the defaults
    DeepTableCopy(settings, defaults);
    DeepTableCopy(defaults, settings);
    self.settings = settings;
    self:SetTabText(settings.name);
    self:SetTabLeft(settings.tabLeft);
    
    self.table = Thurallor.UI.TableControl();
    self.table:SetParent(self);
    self.table:AddColumn("desc", "Description", 2, nil);
    self.table:AddColumn("category", "Category", 1, nil);
    self.table:AddColumn("charName", "Character", 1, nil);
    local function remainingSort(a, b)
        return a.sortValue < b.sortValue;
    end
    self.table:AddColumn("remaining", "Time", 1, remainingSort);
    self.table:SetPresentation(self.settings.presentation);
    AddCallback(self.table, "PresentationChanged", function(_, args)
        self.settings.presentation = self.table:GetPresentation();
        self:SaveSettings();
    end);
    AddCallback(self.table, "SelectionChanged", function(_, args)
        self:SelectionChanged(args);
    end);

    self:SetInteriorControl(self.table);
    self:SetInteriorAlignment(Turbine.UI.ContentAlignment.TopLeft);
    
    self.x_icon = self:CreateIcon("x.tga", "x_highlight.tga");
    self.x_icon:SetPosition(1, 1);
    Thurallor.UI.Tooltip("Delete this reminder"):Attach(self.x_icon);
    self.reset_icon = self:CreateIcon("reset.tga", "reset_highlight.tga");
    self.reset_icon:SetPosition(23, 1);
    local delay = self.window.settings.postponeDelay;
    Thurallor.UI.Tooltip(function() return "Postpone this for " .. self.window.settings.postponeDelay .. " hours"; end):Attach(self.reset_icon);
    
    self.events = {};
    self:SendToBack();
    self:CreateSavedEvents();
end

function Tab:ShowTabContextMenu()
    local contextMenu = Turbine.UI.ContextMenu();
    local menuItems, item = contextMenu:GetItems();

    item = Turbine.UI.MenuItem("Create new tab");
    item.Click = function()
        self.window:CreateNewTab();
    end
    menuItems:Add(item);

    item = Turbine.UI.MenuItem("Rename this tab");
    item.Click = function()
        self:OpenRenameDialog();
    end
    menuItems:Add(item);

    local item = Turbine.UI.MenuItem("Clone this tab");
    item.Click = function()
        self.window:CloneTab(self.settings.name);
    end
    menuItems:Add(item);

    local item = Turbine.UI.MenuItem("Delete this tab");
    item.Click = function()
        self.window:DeleteTab(self.settings.name);
    end
    menuItems:Add(item);
    if (self.window:GetNumUserTabs() < 2) then
        item:SetEnabled(false);
    end

    contextMenu:ShowMenu();
end

function Tab:ShowEventContextMenu(event)
    local contextMenu = Turbine.UI.ContextMenu();
    local menuItems, item = contextMenu:GetItems();


    item = Turbine.UI.MenuItem("Move to other tab");
    item.Click = function()
        self.window:CreateNewTab();
    end
    menuItems:Add(item);
    if (self.window:GetNumUserTabs() == 1) then
        item:SetEnabled(false);
    else
        -- Submenu with a list of tabs
        local subMenuItems = item:GetItems();
        local tabNames = self.window:GetUserTabs();
        for _, name in ipairs(tabNames) do
            if (name ~= self.settings.name) then
                local subItem = Turbine.UI.MenuItem(name);
                subItem.Click = function()
                    local otherTab = self.window:GetTab(name);
                    otherTab:CloneEvent(event);
                    self:DeleteEvent(event);
                end
                subMenuItems:Add(subItem);
            end
        end
    end

    item = Turbine.UI.MenuItem("Copy to other tab");
    item.Click = function()
        self.window:CreateNewTab();
    end
    menuItems:Add(item);
    if (self.window:GetNumUserTabs() == 1) then
        item:SetEnabled(false);
    else
        -- Submenu with a list of tabs
        local subMenuItems = item:GetItems();
        local tabNames = self.window:GetUserTabs();
        for _, name in ipairs(tabNames) do
            if (name ~= self.settings.name) then
                local subItem = Turbine.UI.MenuItem(name);
                subItem.Click = function()
                    local otherTab = self.window:GetTab(name);
                    otherTab:CloneEvent(event);
                end
                subMenuItems:Add(subItem);
            end
        end
    end
    
    local item = Turbine.UI.MenuItem("Clone");
    item.Click = function()
        self:CloneEvent(event);
    end
    menuItems:Add(item);

    local item = Turbine.UI.MenuItem("Delete");
    item.Click = function()
        self:DeleteEvent(event);
    end
    menuItems:Add(item);

    local item = Turbine.UI.MenuItem("Add 24 hours");
    item.Click = function()
        event:SetExpirationTime(event:GetExpirationTime() + 24 * 60 * 60);
    end
    menuItems:Add(item);
    
    contextMenu:ShowMenu();
end

function Tab:SetTabLeft(left)
    Thurallor.UI.TabCard.SetTabLeft(self, left);
    self.settings.tabLeft = left;
    self:SaveSettings();
end

function Tab:Rename(newName)
    if (newName ~= self.settings.name) then
        self.window:RenameTab(self.settings.name, newName);
    end    
end

function Tab:OpenRenameDialog()
    _G.Turbine.UI.Lotro.Action.Accept_Input = 162; -- KeyDown Action arg when user presses Enter key
    local win = Turbine.UI.Lotro.Window();
    win:SetText("Rename");
    win:SetSize(200, 100);
    win:SetVisible(true);
    win:SetZOrder(self.window:GetZOrder() + 1);
    local textBox = Turbine.UI.Lotro.TextBox();
    textBox:SetParent(win);
    textBox:SetPosition(5, 44);
    textBox:SetSize(190, 24);
    textBox:SetFont(Turbine.UI.Lotro.Font.Verdana14);
    textBox:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
    textBox:SetMultiline(false);
    textBox:SetText(self.settings.name);
    textBox:SelectAll();
    textBox:Focus();
    win:Activate();
    textBox:SetWantsKeyEvents(true);
    textBox.KeyDown = function(tb, args)
        if (args.Action == Turbine.UI.Lotro.Action.Accept_Input) then
            local newName = tb:GetText();
            if ((newName == "") or (self.window:GetTab(newName))) then
                -- Unacceptable.
                tb:SetText(self.settings.name);
                tb:SelectAll();
                return;
            end
            win:Close();
            self:Rename(newName);
        end
    end
    self.window:WaitForChild(win);
    -- Position the window over the mouse cursor:
    self.window.mask:MouseClick();
end

function Tab:CreateIcon(normalImage, highlightImage)
    local icon = Turbine.UI.Control();
    icon:SetBlendMode(Turbine.UI.BlendMode.Overlay);
    icon:SetSize(22, 22);
    icon.bgNormal = Turbine.UI.Graphic(imagePath .. normalImage);
    icon.bgHighlight = Turbine.UI.Graphic(imagePath .. highlightImage);
    
    function icon.IsMouseOver(ctl)
        local x, y = ctl:GetMousePosition();
        return ((x >= 0) and (y >= 0) and (x < 22) and (y < 22));
    end
    
    AddCallback(icon, "MouseEnter", function(ctl)
        ctl:SetBackground(ctl.bgHighlight);
    end);
    AddCallback(icon, "MouseLeave", function(ctl)
        ctl:SetBackground(ctl.bgNormal);
    end);
    
    return icon;
end


function Tab:SizeChanged()
    self.table:SetSize(self:GetWidth() - 10, self:GetHeight() - 30);
end

function Tab:TabLeftChanged(tabLeft)
    self.settings.tabLeft = tabLeft;
    self:GetParent():RedistributeTabs(); -- includes save
end

function Tab:CreateEvent(desc, charName, expTime, category)
    local settings = {};
    local event = Event(self, settings, desc, charName, expTime, category);
    local row, cells = self.table:AddRow(event);
    event:SetCells(cells);
    table.insert(self.settings.events, settings);
    self.events[event] = settings;
    
    -- Add "X" icon to delete row, when mouse is present
    AddCallback(row, "MouseEnter", function(ctl)
        self.x_icon:SetBackground(self.x_icon.bgNormal);
        self.x_icon:SetParent(ctl);
        self.x_icon:SetZOrder(4); -- in front of mask
        self.x_icon.MouseClick = function()
            self:DeleteEvent(event);
        end
        if (event.expired) then
            self.reset_icon:SetBackground(self.reset_icon.bgNormal);
            self.reset_icon:SetParent(ctl);
            self.reset_icon:SetZOrder(4); -- in front of mask
            self.reset_icon.MouseClick = function(ctl)
                local delay = self.window.settings.postponeDelay;
                local expTime = Turbine.Engine.GetGameTime() + delay * 60 * 60;
                event:SetExpirationTime(expTime);
                event:SaveSettings();
                ctl:SetParent(nil);
            end
        end

        -- Turn off shimmering
        if (ctl.shimmer) then
            ctl.shimmer = nil;
            for c = 1, #ctl.cells do
                self.window.shimmering:Remove(ctl.cells[c]);
            end
        end
    end);
    AddCallback(row, "MouseLeave", function(ctl)
        if (not (self.x_icon:IsMouseOver() or self.reset_icon:IsMouseOver())) then
            self.x_icon:SetParent(nil);
            self.reset_icon:SetParent(nil);
        end
    end);
    row.ContextMenu = function()
        self:ShowEventContextMenu(event);
    end
    
    self.table:EnsureVisible(event);
    self.table:SelectRow(event);
    row:Focus();
end

function Tab:CreateSavedEvents()
    local oldSettings = self.settings.events;
    self.settings.events = {};
    for _, s in pairs(oldSettings) do
        self:CreateEvent(s.desc, s.charName, s.expTime, s.category);
    end
end

function Tab:CloneEvent(event)
    local desc = event:GetDescription();
    local charName = event:GetCharacterName();
    local expTime = event:GetExpirationTime();
    local category = event:GetCategory();
    self:CreateEvent(desc, charName, expTime, category);
    self:ReapplySort();
end

function Tab:DeleteEvent(event)
    self.table:DeleteRow(event);
    local settings = self.events[event];
    self.events[event] = nil;
    for n, s in pairs(self.settings.events) do
        if (s == settings) then
            self.settings.events[n] = nil;
            break;
        end
    end
    event:Destroy();
    self:SaveSettings();
end

function Tab:SaveSettings(now)
    self.window:SaveSettings(now);
end

-- Function must accept a single 'event' argument.
-- Returning nil will continue the loop.
-- Returning 'true' aborts the loop and returns the event object.
function Tab:ForEachEventDo(func)
    for event, _ in pairs(self.events) do
        if (func(event)) then
            return event;
        end
    end
end

function Tab:Tick(lastTickTime, currentTime)
    -- Call Tick() function of each event once per second.
    -- To do: Make this smarter, to spread out the event updates over multiple frame updates.
    -- Probably no need to update as often events that aren't going to happen for a long time.
    for event, settings in pairs(self.events) do
        event:Tick();
    end
end

function Tab:EventExpired(event)
    local controls = {};

    -- Make the event shimmer in the table
    local cells = event:GetCells();
    cells["desc"]:GetParent().shimmer = true;
    for colName, cell in pairs(cells) do
        table.insert(controls, cell);
    end
    
    -- If the tab is not in front, make the tab shimmer as well
    if (not self:IsInFront()) then
        table.insert(controls, self.tabText);
    end

    -- If the window is iconified, make the icon shimmer as well
    local desc = cells["desc"]:GetText();
    self.window:AddNotification(desc, controls);
end

function Tab:Deselect()
    self.table:Deselect()
end

function Tab:CleanupDropDowns()
    if (self.dropDownCtl) then
        self.dropDownCtl.dropDown:SetParent(nil);
        self.dropDownCtl.dropDown = nil;
        self.dropDownCtl = nil;
    end
end

function Tab:SelectionChanged(args)
    self:CleanupDropDowns();
    if (args.Old) then
        args.Old:SelectionLost();
    end
    if (args.New) then
        args.New:SelectionGained();
    end
end

-- Resorts the table if and only if it's sorted on the specified column
function Tab:ReapplySort(columnName)
    if (self.table:IsSorted(columnName)) then
        self.table:ReSort();
    end
end
